-------------------------------------------------------------------------------
--! @file       CipherCore.vhd
--! @author     Farnoud Farahmand
--! @brief      AES_COPA CipherCore
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.ALL;

entity CipherCore is
    generic (
        --! Reset behavior
        G_ASYNC_RSTN    : boolean := False; --! Async active low reset
        --! Block size (bits)
        G_DBLK_SIZE     : integer := 128;   --! Data
        G_KEY_SIZE      : integer := 128;   --! Key
        G_TAG_SIZE      : integer := 128;   --! Tag
        --! The number of bits required to hold block size expressed in
        --! bytes = log2_ceil(G_DBLK_SIZE/8)
        G_LBS_BYTES     : integer := 4
    );
    port (
        --! Global
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! PreProcessor (data)
        key             : in  std_logic_vector(G_KEY_SIZE       -1 downto 0);
        bdi             : in  std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        --! PreProcessor (controls)
        key_ready       : out std_logic;
        key_valid       : in  std_logic;
        key_update      : in  std_logic;
        decrypt         : in  std_logic;
        bdi_ready       : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_type        : in  std_logic_vector(3                -1 downto 0);
        bdi_partial     : in  std_logic;
        bdi_eot         : in  std_logic;
        bdi_eoi         : in  std_logic;
        bdi_size        : in  std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        bdi_valid_bytes : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        bdi_pad_loc     : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        --! PostProcessor
        bdo             : out std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        bdo_valid       : out std_logic;
        bdo_ready       : in  std_logic;
        bdo_size        : out std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        msg_auth_done   : out std_logic;
        msg_auth_valid  : out std_logic
    );
end entity CipherCore;

architecture structure of CipherCore is


-----------DATAPATH------------------------------------------------	  		  
	    signal  start1,start2 :  std_logic;
	    signal  tag_match     :  std_logic;
        signal  ready1,ready2 :  std_logic;
        signal  done1,done2   :  std_logic;
		signal  done_init_both:  std_logic;
     
-----------Controller-----------------------------------------------------
        signal init_delta     :  std_logic;
   	    signal init_pmac      :  std_logic;
		signal init1,init2    :  std_logic:= '0';
		signal init_Vi        :  std_logic;
	    signal init_Enc_Dec   :  std_logic;
		signal init_enc_dec_tag : std_logic;
		signal init_U         :  std_logic;
		signal init_size_eq   :  std_logic;
		signal init_in1       :  std_logic_vector(2 downto 0);
		signal init_out       :  std_logic;
		signal init_tag           : std_logic;
		signal init_partial_delta : std_logic;
-------------------------------------------------------------------
        signal en_Vi         : std_logic;
		signal en_L          : std_logic;
		signal en_delta1     : std_logic;
		signal en_delta0     : std_logic;
        signal en_Vi_dec     : std_logic;
        signal en_U          : std_logic;
        signal en_input_tag  : std_logic;
        signal en_Enc_Dec    : std_logic;
        signal en_sum        : std_logic;		  
		signal en_S          : std_logic;
		signal rst_sum       : std_logic;
		signal start_len_cal : std_logic;
		signal bdo_valid_sig : std_logic;
		signal decrypt_r_p   : std_logic;

begin

    Datapath: entity work.copa_datapath(basic_arc) 
    port map (
                rst                   => rst,
		        clk                   => clk,
				bdi                   => bdi,
				key                   => key,
				bdo                   => bdo,	  	  
		        start1                => start1,
				start2                => start2,
		        tag_match             => tag_match,
                ready1                => ready1,
				ready2                => ready2,
                done1                 => done1,
				done2                 => done2,
		        done_init_both        => done_init_both,
				init_delta            => init_delta,
		        init_pmac             => init_pmac,
		        init1                 => init1,
				init2                 => init2,
		        init_Vi               => init_Vi,
		        init_Enc_Dec          => init_Enc_Dec,
				init_enc_dec_tag      => init_enc_dec_tag,  
		        init_U                => init_U,
				init_size_eq          => init_size_eq,
		        init_in1              => init_in1,
		        init_out              => init_out,
				init_partial_delta    =>  init_partial_delta,
		        init_tag              => init_tag,
				en_Vi                 => en_Vi,
		        en_L                  => en_L,
		        en_delta1             => en_delta1,
				en_delta0             => en_delta0,
				en_Vi_dec             => en_Vi_dec,
				en_U                  => en_U,
				en_input_tag          => en_input_tag,
				en_Enc_Dec            => en_Enc_Dec,
				en_sum                => en_sum,
		        en_S                  => en_S,
				rst_sum               => rst_sum,
				bdi_valid_bytes	      => bdi_valid_bytes,
			    bdi_pad_loc 		  => bdi_pad_loc,
				decrypt 		      => decrypt_r_p,
				bdo_size			  => bdo_size,
				start_len_cal		  => start_len_cal,
				bdo_valid			  => bdo_valid_sig
				);
				 
    Controller: entity work.copa_controller(Behavioral) 
	 port map (		 
				rst                   => rst,
				clk                   => clk,	  	  
				start1                => start1,
				start2                => start2,
				done1                 => done1,
				done2                 => done2,
				done_init_both        => done_init_both,
				init_delta            => init_delta,
				init_pmac             => init_pmac,
				init1                 => init1,
				init2                 => init2,
				init_Vi               => init_Vi,
				init_Enc_Dec          => init_Enc_Dec,
				init_enc_dec_tag      => init_enc_dec_tag,  
				init_U                => init_U,
				init_size_eq          => init_size_eq,
				init_in1              => init_in1,
				init_out              => init_out,
				init_tag              => init_tag,
				init_partial_delta    =>  init_partial_delta,
				en_Vi                 => en_Vi,
				en_L                  => en_L,
				en_delta1             => en_delta1,
				en_delta0             => en_delta0,
				en_Vi_dec             => en_Vi_dec,
				en_U                  => en_U,
				en_input_tag          => en_input_tag,
				en_Enc_Dec            => en_Enc_Dec,
				en_sum                => en_sum,
				en_S                  => en_S,
				rst_sum               => rst_sum,
				tag_match             => tag_match,
				key_update            => key_update,
				key_valid             => key_valid,
				key_ready             => key_ready,
				bdi_valid             => bdi_valid,
				bdi_ready             => bdi_ready,
				bdi_decrypt           => decrypt,
				bdi_eot               => bdi_eot,
				bdi_partial			  => bdi_partial,
			    bdi_pad_loc 		  => bdi_pad_loc,
				bdo_ready             => bdo_ready,
				bdo_valid             => bdo_valid_sig,
				msg_auth_valid        => msg_auth_valid,
				msg_auth_done         => msg_auth_done,
				start_len_cal		  => start_len_cal,
				decrypt_r_p			  => decrypt_r_p
			 );
			 
			 bdo_valid  <= bdo_valid_sig;


end structure;

