-------------------------------------------------------------------------------
--! @file       copa_controller.vhd
--! @author     Farnoud Farahmand
--! @brief      Control unit for AES_COPA 
-------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

entity copa_controller is
     port(  
          clk       : in std_logic;
          rst       : in std_logic;
			 
          --! Internal
			 -------MUX-INIT---------
           init_delta       : out   std_logic;
		   init_pmac        : out   std_logic;
		   init1,init2      : out   std_logic;
		   init_Vi          : out   std_logic;
		   init_Enc_Dec     : out   std_logic;
		   init_U           : out   std_logic;
		   init_size_eq     : out   std_logic;
		   init_Enc_Dec_tag : out   std_logic;
		   init_in1         : out   std_logic_vector(2 downto 0);
		   init_out         : out   std_logic;
		   init_tag         : out   std_logic;
		   init_partial_delta : out   std_logic;
		   
         --------EN-----------		  
		   en_Vi            : out   std_logic;
		   en_L             : out   std_logic;
		   en_delta1        : out   std_logic;
		   en_delta0        : out   std_logic;
           en_U             : out   std_logic;
           en_input_tag     : out   std_logic;
           en_sum           : out   std_logic;		  
		   en_S             : out   std_logic;			
		   en_Enc_Dec       : out   std_logic;
		   en_Vi_dec        : out   std_logic;
		   rst_sum          : out   std_logic;
		   tag_match        : in    std_logic;
			--------AES-----------
		   start1,start2    : out   std_logic;
           done1,done2      : in  std_logic;
		   done_init_both   : in  std_logic;
			
			--! External
			key_valid        : in  std_logic;
			key_ready        : out std_logic;
			key_update       : in  std_logic;
			bdi_valid        : in  std_logic;
			bdi_ready        : out std_logic;
			bdi_decrypt      : in  std_logic;
			bdi_eot          : in  std_logic;
			bdi_partial		 : in  std_logic;	
			bdo_ready       : in  std_logic;
			bdo_valid       : out std_logic;
			msg_auth_valid   : out std_logic;
			msg_auth_done    : out std_logic;
			bdi_pad_loc   : in std_logic_vector(15 downto 0);
			start_len_cal : out std_logic;
			decrypt_r_p	  : out std_logic
			);
			
end copa_controller;

architecture Behavioral of copa_controller is

      --! Internal Registers
      type t_state is (S_RESET, S_KEY_READ, S_KEY_SCH, S_CAL_L, S_AD_INIT,
                  	  S_AD_ENC, S_AD_LASTBLOCK, S_M_ENC, S_M_DEC, S_ENC_DEC,
							  S_WR_MC, S_DEC_LAST, S_WR_LASTM, S_TAG_READ, S_SUM_ENC,
							  S_TAG_ENC, S_WR_TAG, S_WR_LASTME, S_DELTA_INIT, 
							  S_WAIT, S_INIT_ENC, S_INIT_DEC, S_TAG_DEC
							  );
							  
		signal state            : t_state;
        signal state_next       : t_state;
		signal decrypt_r        : std_logic;
		signal lastbl_r         : std_logic;
		signal partial_r         : std_logic;
		signal firstblock_r     : std_logic;
		signal firstblock_next  : std_logic;		
		signal enc_dec_r        : std_logic;
		signal enc_dec_next     : std_logic;
		signal enc_dec_tag_r    : std_logic;
		signal enc_dec_tag_next : std_logic;
		
      --! Internal signals
        signal en_decrypt_s     : std_logic;
		signal en_lastbl_s      : std_logic;
		signal en_partial_s      : std_logic;		
		
begin
   
	 p_fsm: process(clk)
     begin
        if rising_edge(clk) then
          if (rst = '1') then
            state   <= S_RESET;
          else
            state   <= state_next;
          end if;
			 
          if (en_decrypt_s = '1') then
            decrypt_r <= bdi_decrypt;
          end if;
          if (en_lastbl_s = '1') then
            lastbl_r <= bdi_eot;
          end if;
			 if (en_partial_s = '1') then
            partial_r <= bdi_partial;
          end if;
			 firstblock_r  <= firstblock_next;
			 enc_dec_r     <= enc_dec_next;
			 enc_dec_tag_r <= enc_dec_tag_next;
        end if; 
      end process;
		
		init_Enc_Dec      <= enc_dec_r;
		init_Enc_Dec_tag  <= enc_dec_tag_r;
		decrypt_r_p		  <= decrypt_r;	
		
     p_comb: process(state, done1, done2, done_init_both,  key_update,
	                  key_valid, bdi_valid, decrypt_r, bdi_decrypt, bdi_eot,
							lastbl_r, bdi_pad_loc, bdo_ready, firstblock_r, enc_dec_r, enc_dec_tag_r, partial_r)
	    begin
          --! Default values
			state_next         <= state;
            init_delta   <= '0';
            init_pmac    <= '0';
            init1        <= '0';
            init2        <= '0';
            init_Vi      <= '0';
            init_U       <= '0';
            init_in1     <= "000";
            init_out     <= '0';
			init_size_eq <= '0';
            en_Vi        <= '0';
			en_Vi_dec    <= '0';
			en_Enc_Dec   <= '0';
            en_L         <= '0';
            en_delta1    <= '0';
		    en_delta0    <= '0';
            en_U         <= '0';
            en_input_tag <= '0';
            en_sum       <= '0';
            en_S         <= '0';
			rst_sum      <= '0';
			start1       <= '0';
            start2       <= '0';
            key_ready    <= '0';
            bdi_ready    <= '0';
		    bdo_valid    <= '0';
			en_decrypt_s <= '0';
			en_lastbl_s  <= '0';
			bdo_valid    <= '0';
			msg_auth_valid  <= '0';
			msg_auth_done   <= '0';
			init_tag        <= '0';
			en_partial_s    <= '0';
			init_partial_delta  <= '0';
			start_len_cal       <= '0';
			
		case state is
		        when S_RESET =>
					if (bdi_valid = '1')then
					    if (key_update = '1')then
                    	    state_next <= S_KEY_READ;
                    	else
                        	init_pmac  <= '1';
							init_delta <= '1';
							en_delta0  <= '1';
							init_u     <= '1';
							en_u       <= '1';
							rst_sum    <= '1';
							state_next <= S_AD_INIT;
						end if; 					  
			        end if;
				
				when S_KEY_READ => 
					key_ready <= '1';
					if (key_valid = '1')then
						init1 <='1';
					    init2 <='1';
						state_next <= S_KEY_SCH;
			        end if;
				 
				when S_KEY_SCH =>
					enc_dec_next <= '0';
				    if (Done_init_both = '1')then
					    start1     <='1';
                    	state_next <= S_CAL_L;				  
			        end if;
				
				when S_CAL_L =>
				    if (done1 = '1')then
					    en_L       <='1';
                    	state_next <= S_DELTA_INIT;				  
			        end if;
						 
				when S_DELTA_INIT =>	   
					    init_pmac  <= '1';
					    init_delta <= '1';
						en_delta0  <= '1';
					    init_u     <= '1';
						en_u       <= '1';
						rst_sum    <= '1';
						enc_dec_next <= '0';
						state_next <= S_AD_INIT;
				  
				when S_AD_INIT =>
						bdi_ready <= '1';
						if(bdi_valid = '1')then
                            if(bdi_eot = '1')then
                            	if(bdi_pad_loc = x"0000")then							  
							            start1    <= '1';
							            init_in1  <= "110";
							            bdi_ready  <= '1';
							            state_next <= S_AD_LASTBLOCK; 
							    else
										init_size_eq  <= '1';
										start1        <= '1';
							            init_in1      <= "110";
							            bdi_ready     <= '1';
							            state_next    <= S_AD_LASTBLOCK; 
								end if;
						    else		  
							    start1      <= '1';
							    init_in1    <= "101";
							    bdi_ready   <= '1'; 
							    state_next  <= S_AD_ENC;
							end if; 
					    end if;		 							 
							
				when S_AD_ENC =>
					if (done1 = '1')then
						en_u       <= '1';
						en_delta0  <= '1';
						state_next <= S_AD_INIT;
					end if; 
					   
				when S_AD_LASTBLOCK =>
					if (done1 = '1')then	 
							init_delta   <= '1';
							en_delta0    <= '1';
							en_delta1    <= '1';
							init_Vi      <= '1';
						    en_Vi        <= '1';
							en_vi_dec    <= '1';
                    	    en_decrypt_s <= '1';
                            firstblock_next <= '1';
							en_partial_s    <= '1'; --?
                         if (bdi_decrypt = '1')then	
                            state_next <= S_INIT_DEC;
                         else
                            state_next <= S_INIT_ENC;
                         end if;
					end if;

                when S_INIT_ENC =>
                        enc_dec_next     <= '0';
					    enc_dec_tag_next <= '0';
                        state_next <= S_M_ENC;

                when S_INIT_DEC =>
                        enc_dec_next     <= '1';
					    enc_dec_tag_next <= '1';
                        state_next <= S_M_DEC;								
                   
                when S_M_ENC =>
						bdi_ready        <= '1';
						init_in1         <= "001";
					    if (bdi_valid = '1')then
								if (firstblock_r = '1')then
										start1      <= '1';
										en_lastbl_s <= '1';
										en_partial_s <= '1';
										en_sum      <= '1';						 
										state_next  <= S_ENC_DEC;
								else
										if ((bdi_partial = '1') and (bdi_eot = '1'))then
												init_partial_delta <= '1';
												start1      <= '1';
												en_lastbl_s <= '1';
												en_partial_s <= '1';
												en_sum      <= '1';						 
												state_next  <= S_ENC_DEC;
										else
												start1      <= '1';
												en_lastbl_s <= '1';
												en_partial_s <= '1';
												en_sum      <= '1';						 
												state_next  <= S_ENC_DEC;
										end if;
								end if;
                        end if;
							
				when S_M_DEC =>
						bdi_ready        <= '1';
						init_in1         <= "010";
						if (bdi_valid = '1')then	 
								start1       <= '1';
								en_lastbl_s  <= '1';	
								en_partial_s <= '1';		
								state_next   <= S_ENC_DEC;
				        end if;
							
				when S_ENC_DEC=>
					if (done1 = '1')then
									if (firstblock_r = '1')then
													firstblock_next  <= '0';
													en_Vi      <= '1';
													en_enc_dec <= '1';
													start2     <= '1';	
													if (lastbl_r = '1')then
																if (decrypt_r = '1')then
																		state_next  <= S_DEC_LAST;
																else	
																		if (partial_r = '1')then
																				init_partial_delta <= '1';
																				Start1             <= '1';
																				init_in1           <= "011";
																				en_s               <= '1';
																				state_next         <= S_SUM_ENC;
																		else
																				Start1             <= '1';
																				init_in1           <= "011";
																				en_s               <= '1';
																				state_next         <= S_SUM_ENC;
																		end if;
																end if;	
													else
																if (decrypt_r = '1')then
																		en_delta1    <= '1';	
																		state_next   <= S_M_DEC;
																else
																		en_delta0    <= '1';
																		state_next   <= S_M_ENC;
																end if;					
													end if;		
									else
													if ( bdo_ready = '1')then  
															    if (decrypt_r = '1')then
																			bdo_valid    <= '1';
																			init_out     <= '1';
																			en_sum       <= '1';
																			en_delta0    <= '1';
																			en_Vi        <= '1';
																			en_enc_dec   <= '1';
																			start2       <= '1';	
																			if (lastbl_r = '1')then
																						if (decrypt_r = '1')then
																								state_next  <= S_DEC_LAST;
																						else	
																								if (partial_r = '1')then
																										init_partial_delta <= '1';
																										Start1             <= '1';
																										init_in1           <= "011";
																										en_s               <= '1';
																										state_next         <= S_SUM_ENC;
																								else
																										Start1             <= '1';
																										init_in1           <= "011";
																										en_s               <= '1';
																										state_next         <= S_SUM_ENC;
																								end if; 
																						end if;	
																			else
																						if (decrypt_r = '1')then
																								en_delta1    <= '1';	
																								state_next   <= S_M_DEC;
																						else
																								en_delta0    <= '1';
																								state_next   <= S_M_ENC;
																						end if;					
																			end if;											  
																else
																			bdo_valid   <= '1';
																			en_delta1    <= '1';
																			en_Vi        <= '1';
																			en_enc_dec   <= '1';
																			start2       <= '1';	
																			if (lastbl_r = '1')then
																						if (decrypt_r = '1')then
																								state_next  <= S_DEC_LAST;
																						else	
																								if (partial_r = '1')then
																										init_partial_delta <= '1';
																										Start1             <= '1';
																										init_in1           <= "011";
																										en_s               <= '1';
																										state_next         <= S_SUM_ENC;
																								else
																										Start1             <= '1';
																										init_in1           <= "011";
																										en_s               <= '1';
																										state_next         <= S_SUM_ENC;
																								end if; 
																						end if;	
																			else
																						if (decrypt_r = '1')then
																								en_delta1    <= '1';	
																								state_next   <= S_M_DEC;
																						else
																								en_delta0    <= '1';
																								state_next   <= S_M_ENC;
																						end if;					
																			end if;
																end if;	
													else		
															    state_next   <= S_WR_MC;			
													end if;				
									end if;
					else
									if (firstblock_r = '1')then
											state_next   <= S_ENC_DEC;
									else
											en_vi_dec    <= '1';
											state_next   <= S_ENC_DEC;
									end if;											
				    end if;
               
                when S_WR_MC=>					
						if ( bdo_ready = '1')then  
								if (decrypt_r = '1')then
											bdo_valid    <= '1';
											init_out     <= '1';
											en_sum       <= '1';
											en_delta0    <= '1';
											en_Vi        <= '1';
											en_enc_dec   <= '1';
											start2       <= '1';	
											if (lastbl_r = '1')then
														if (decrypt_r = '1')then
																state_next  <= S_DEC_LAST;
														else	
																if (partial_r = '1')then
																		init_partial_delta <= '1';
																		Start1             <= '1';
																		init_in1           <= "011";
																		en_s               <= '1';
																		state_next         <= S_SUM_ENC;
																else
																		Start1             <= '1';
																		init_in1           <= "011";
																		en_s               <= '1';
																		state_next         <= S_SUM_ENC;
																end if; 
														end if;	
											else
														if (decrypt_r = '1')then
																en_delta1    <= '1';	
																state_next   <= S_M_DEC;
														else
																en_delta0    <= '1';
																state_next   <= S_M_ENC;
														end if;					
											end if;											  
								else
											bdo_valid   <= '1';
											en_delta1    <= '1';
											en_Vi        <= '1';
											en_enc_dec   <= '1';
											start2       <= '1';	
											if (lastbl_r = '1')then
														if (decrypt_r = '1')then
																state_next  <= S_DEC_LAST;
														else	
																if (partial_r = '1')then
																		init_partial_delta <= '1';
																		Start1             <= '1';
																		init_in1           <= "011";
																		en_s               <= '1';
																		state_next         <= S_SUM_ENC;
																else
																		Start1             <= '1';
																		init_in1           <= "011";
																		en_s               <= '1';
																		state_next         <= S_SUM_ENC;
																end if;
														end if;	
											else
														if (decrypt_r = '1')then
																en_delta1    <= '1';	
																state_next   <= S_M_DEC;
														else
																en_delta0    <= '1';
																state_next   <= S_M_ENC;
														end if;					
											end if;
								end if;		
						end if;
					
                when S_DEC_LAST=>
						en_vi_dec  <= '1';
						if (done2 = '1')then
								if ( bdo_ready = '1')then
										if (partial_r = '1')then
												init_partial_delta <= '1'; 
												en_sum        <= '1';
												start_len_cal <= '1';
												state_next   <= S_WAIT;
										else
												init_out     <= '1';
												en_sum       <= '1';
												bdo_valid    <= '1'; 
												state_next   <= S_WAIT;
										end if;
								else
										state_next   <= S_WR_LASTM; 											  
								end if;
                        end if;
								  
                when S_WR_LASTM=>
                        if ( bdo_ready = '1')then
										if (partial_r = '1')then
												init_partial_delta <= '1'; 
												en_sum        <= '1';
												start_len_cal <= '1';
												state_next   <= S_WAIT;
										else
												init_out     <= '1';
												en_sum       <= '1';
												bdo_valid    <= '1'; 
												state_next   <= S_WAIT;
										end if;
						end if;
									 
				when S_WAIT=>
						enc_dec_next <= '0';
                        state_next   <= S_TAG_READ;						

                when S_TAG_READ=>
						if (partial_r = '1')then
								init_partial_delta <= '1';
								Start1             <= '1';
								init_in1           <= "011";
								en_s               <= '1';
                                state_next         <= S_SUM_ENC;
						else
								Start1             <= '1';
								init_in1           <= "011";
								en_s               <= '1';
                                state_next         <= S_SUM_ENC;
						end if;
									 
				when S_SUM_ENC=>			 
					    if (done1 = '1')then
							    Start1        <= '1';
								init_in1      <= "100";
								if (decrypt_r = '1')then									
										state_next         <= S_TAG_ENC;
								else
								        if ( bdo_ready = '1')then
										    bdo_valid     <= '1';
											state_next   <= S_TAG_ENC;
										else		  
											state_next   <= S_WR_LASTME;
								        end if;
						        end if;														
						end if;
									  
				when S_WR_LASTME=>
				        if ( bdo_ready = '1')then
							    bdo_valid     <= '1';
							    state_next   <= S_TAG_ENC;	
				        end if;											
									  
				when S_TAG_ENC=>		
							enc_dec_next <= '1';
					        init_tag    <= '1';	  
							if (done1 = '1')then
									if (decrypt_r = '1')then
											if (partial_r = '1')then
													init_partial_delta  <= '1';
													init_out			<= '1';
													bdo_valid			<= '1';
													state_next    <= S_TAG_DEC;	
											else
													state_next    <= S_TAG_DEC;	
											end if;
									else    								
											if (bdo_ready = '1')then 
												    bdo_valid    <= '1';
													state_next   <= S_RESET;
											else
													state_next   <= S_WR_TAG;
											end if; 																		  
									end if;			  
							end if;	

                when S_WR_TAG=>	
                        init_tag    <= '1';       						  
                        if (bdo_ready = '1')then 
								bdo_valid    <= '1';
								state_next   <= S_RESET;
						end if;
									  
				when S_TAG_DEC=>	
					    bdi_ready    <= '1';       						  
						if (bdi_valid = '1')then 
								if (tag_match = '1')then
										msg_auth_valid  <= '1';
										msg_auth_done   <= '1';
										state_next      <= S_RESET;
								else
										msg_auth_done   <= '1';
										state_next      <= S_RESET;
								end if;
						end if;
          
			end case;
		end process;
				
end Behavioral;                               


